#include<iostream>
#include<vector>
#include<limits>
#include<cmath>
#include<cassert>

template<class T>
void F(int bw, const std::vector<T>& x, std::vector<T> &y) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  for (int i=0;i<n;i++) {
    y[i]=1;
    for (int j=std::max(0,i-bw/2);j<=std::min(n-1,i+bw/2);j++) y[i]*=sin(x[j]);
  }
}

#include "dco.hpp"

template<class T>
void dFdx(
    int bw,
    std::vector<T> x,
    std::vector<std::vector<T>> &dydx
) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  using DCO_M=typename dco::ga1s<T>;
  using DCO_T=typename DCO_M::type;
  using DCO_TT=typename DCO_M::tape_t;
  std::vector<DCO_T> x_a(n), y_a(n);
  dco::value(x_a)=x;
  DCO_M::global_tape=DCO_TT::create();
  DCO_M::global_tape->register_variable(x_a);
  F(bw,x_a,y_a); 
  for (int j=0;j<n;j++) {
    dco::derivative(y_a[j])=1;
    DCO_M::global_tape->interpret_adjoint();
    for (int i=0;i<n;i++) dydx[j][i]=dco::derivative(x_a[i]);
    DCO_M::global_tape->zero_adjoints();
  }
  DCO_TT::remove(DCO_M::global_tape);
}

template<class T>
void dFdx_CPR(
    int bw,
    const std::vector<T> &x,
    std::vector<std::vector<T>> &dydx
) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  using DCO_M=typename dco::ga1s<T>;
  using DCO_T=typename DCO_M::type;
  using DCO_TT=typename DCO_M::tape_t;
  std::vector<DCO_T> x_a(n), y_a(n);
  dco::value(x_a)=x;
  DCO_M::global_tape=DCO_TT::create();
  DCO_M::global_tape->register_variable(x_a);
  F(bw,x_a,y_a); 
  for (int k=0;k<bw;k++) {
    for (int i=k;i<n;i+=bw) dco::derivative(y_a[i])=1;
    DCO_M::global_tape->interpret_adjoint();
    for (int i=0;i<n;i++) dydx[i][k]=dco::derivative(x_a[i]);
    DCO_M::global_tape->zero_adjoints();
  }
  DCO_TT::remove(DCO_M::global_tape);
}

template<class T>
void dFdx_CPR_unpack(
  const std::vector<std::vector<T>> &dydx_CPR,
  std::vector<std::vector<T>> &dydx
) {
  int n=dydx_CPR.size(), bw=dydx_CPR[0].size();
  assert(n%bw==0); assert(bw%2==1); 
  for (int i=0;i<n;i++) 
    for (int j=std::max(0,i-bw/2);j<=std::min(n-1,i+bw/2);j++) 
      dydx[i][j]=dydx_CPR[i][j%bw];
}

int main(int argc, char *argv[]) {
  assert(argc==3); 
  int n=std::stoi(argv[1]), bw=std::stoi(argv[2]);
  using T=double;
  std::vector<T> x(n,1);
{ // dense
  std::vector<std::vector<T>> dydx(n,std::vector<T>(n));
  dFdx(bw,x,dydx);
  for (auto &r : dydx) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
}
{ // CPR
  std::vector<std::vector<T>> dydx_CPR(n,std::vector<T>(bw));
  std::vector<std::vector<T>> dydx(n,std::vector<T>(n,0));
  dFdx_CPR(bw,x,dydx_CPR);
  for (auto &r : dydx_CPR) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
  dFdx_CPR_unpack(dydx_CPR,dydx);
  for (auto &r : dydx) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
}
  return 0;
}
