#include<iostream>
#include<vector>
#include<limits>
#include<cmath>
#include<cassert>
using namespace std;

template<class T>
void F(int bw, const std::vector<T>& x, std::vector<T> &y) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  for (int i=0;i<n;i++) {
    y[i]=1;
    for (int j=max(0,i-bw/2);j<=min(n-1,i+bw/2);j++) y[i]*=sin(x[j]);
  }
}

template<class T>
void dFdx(
    int bw,
    std::vector<T> x,
    std::vector<std::vector<T>> &dydx
) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  std::vector<T> y(n),yp(n);
  F(bw,x,y); 
  for (int i=0;i<n;i++) {
    auto dx=(x[i]==0) ? sqrt(std::numeric_limits<T>::epsilon())
                      : sqrt(std::numeric_limits<T>::epsilon())*fabs(x[i]);
    x[i]+=dx; F(bw,x,yp); x[i]-=dx;
    for (int j=0;j<n;j++) dydx[j][i]=(yp[j]-y[j])/dx;
  }
}

template<class T>
void dFdx_CPR(
    int bw,
    const std::vector<T> &x,
    std::vector<std::vector<T>> &dydx
) {
  int n=x.size(); assert(n%bw==0); assert(bw%2==1); 
  std::vector<T> y(n);
  F(bw,x,y);
  for (int k=0;k<bw;k++) {
    std::vector<T> dx(n/bw), xp=x, yp(n);
    for (int i=k,j=0;i<n;i+=bw,j++) {
      dx[j]=(x[i]==0) ? sqrt(std::numeric_limits<T>::epsilon())
                      : sqrt(std::numeric_limits<T>::epsilon())*fabs(x[i]);
      xp[i]+=dx[j]; 
    }
    F(bw,xp,yp); 
    for (int i=0;i<n;i++) dydx[i][k]=(yp[i]-y[i])/dx[i%(n/bw)];
  }
}

template<class T>
void dFdx_CPR_unpack(
  const std::vector<std::vector<T>> &dydx_CPR,
  std::vector<std::vector<T>> &dydx
) {
  int n=dydx_CPR.size(), bw=dydx_CPR[0].size();
  assert(n%bw==0); assert(bw%2==1); 
  for (int i=0;i<n;i++) 
    for (int j=max(0,i-bw/2);j<=min(n-1,i+bw/2);j++) 
      dydx[i][j]=dydx_CPR[i][j%bw];
}

int main(int argc, char *argv[]) {
  assert(argc==3); 
  int n=std::stoi(argv[1]), bw=std::stoi(argv[2]);
  using T=double;
  std::vector<T> x(n,1);
{ // dense
  std::vector<std::vector<T>> dydx(n,std::vector<T>(n));
  dFdx(bw,x,dydx);
  for (auto &r : dydx) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
}
{ // CPR
  std::vector<std::vector<T>> dydx_CPR(n,std::vector<T>(bw));
  std::vector<std::vector<T>> dydx(n,std::vector<T>(n,0));
  dFdx_CPR(bw,x,dydx_CPR);
  for (auto &r : dydx_CPR) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
  dFdx_CPR_unpack(dydx_CPR,dydx);
  for (auto &r : dydx) {
    for (auto &c : r) std::cout << c << " ";
    std::cout << std::endl;
  }
}
  return 0;
}
